-- Some notes regarding the mechanics and events triggered on weapon enchant changes:
-- * There doesn't seem to be an event for temp weapon enchants, UNIT_INVENTORY_CHANGED works well to determine when you gain or lose one though.
-- * When applying a shaman imbue, a UNIT_AURA event is also triggered unlike non-spell enchants like poisons etc.
-- * The Blizz UI uses an OnUpdate handler to query GetWeaponEnchantInfo() on each update, I don't like that method at all.
-- * Weapon enchant data  seems to be updated *exactly* after the UNIT_INVENTORY_CHANGED event, which is why you cannot update the data during this event.
-- * Zoning seems to "suspend" the weapon buff timer, probably the same mechanics that "suspends" them while being logged out. Due to this, and how new weapon enchants are detected, timers might reset on long loads.

if (type(LibWeaponEnchant) == "table") and (LibWeaponEnchant.vers >= 1) then
	return;
end
LibWeaponEnchant = { vers = 1 };

local f = CreateFrame("Frame");
local callbacks = {};
local nextUpdate;

-- Temp Weapon Enchant Variables
local mhSlot = GetInventorySlotInfo("MainHandSlot");
local ohSlot = GetInventorySlotInfo("SecondaryHandSlot");
local mhEnchant, ohEnchant;
local mhEndTime, ohEndTime = 0, 0;
local mhDuration, ohDuration;
local imbuePattern = "(.+) %(%d+ .+%)";

-- Tooltip
local tip = CreateFrame("GameTooltip","LibWeaponEnchantScannerTip",nil,"GameTooltipTemplate");
tip:SetOwner(UIParent,"ANCHOR_NONE");

--------------------------------------------------------------------------------------------------------
--                                         Scanning Functions                                         --
--------------------------------------------------------------------------------------------------------

-- Scan Tooltip
local function ScanWeaponTipEnchant(slotId)
	tip:ClearLines();
	tip:SetInventoryItem("player",slotId);
	local enchant;
	for i = 6, tip:NumLines() do
		enchant = _G["LibWeaponEnchantScannerTipTextLeft"..i]:GetText():match(imbuePattern);
		if (enchant) then
			return enchant;
		end
	end
	return "??";
end

-- Get Enchant Data
local function QueryWeaponEnchantData()
	local hasChanged;
	local endTime, enchantName;
	local hasMainHandEnchant, mainHandExpiration, _, hasOffHandEnchant, offHandExpiration = GetWeaponEnchantInfo();
	-- Mainhand
	if (hasMainHandEnchant) then
		enchantName = ScanWeaponTipEnchant(mhSlot);
		endTime = GetTime() + (mainHandExpiration / 1000);
		if (mhEnchant ~= enchantName) or (endTime > mhEndTime + 1.2) then
			mhEnchant = enchantName;
			mhEndTime = endTime;
			mhDuration = ceil(mainHandExpiration / 1000);
			hasChanged = true;
		end
	else
		if (mhEnchant) then
			hasChanged = true;
		end
		mhEndTime, mhEnchant = 0, nil;
	end
	-- Offhand
	if (hasOffHandEnchant) then
		enchantName = ScanWeaponTipEnchant(ohSlot);
		endTime = GetTime() + (offHandExpiration / 1000);
		if (ohEnchant ~= enchantName) or (endTime > ohEndTime + 1.2) then
			ohEnchant = enchantName;
			ohEndTime = endTime;
			ohDuration = ceil(offHandExpiration / 1000);
			hasChanged = true;
		end
	else
		if (ohEnchant) then
			hasChanged = true;
		end
		ohEndTime, ohEnchant = 0, nil;
	end
	-- Run Callback Funcs
	if (hasChanged) then
		for token, func in next, callbacks do
			func(token,mhSlot,mhEnchant,mhDuration,mhEndTime,ohSlot,ohEnchant,ohDuration,ohEndTime);
		end
	end
end

--------------------------------------------------------------------------------------------------------
--                                           Updater Scripts                                          --
--------------------------------------------------------------------------------------------------------

-- OnUpdate
local function OnUpdate(self,elapsed)
	nextUpdate = (nextUpdate - elapsed);
	if (nextUpdate <= 0) then
		QueryWeaponEnchantData();
		self:SetScript("OnUpdate",nil);
	end
end

-- OnEvent
local function OnEvent(self,event,unit)
	if (unit == "player" or event ~= "UNIT_INVENTORY_CHANGED") then
		nextUpdate = 0.4;
		self:SetScript("OnUpdate",OnUpdate);
	end
end

f:SetScript("OnEvent",OnEvent);

--------------------------------------------------------------------------------------------------------
--                                 Callback Registration  +  Get Data                                 --
--------------------------------------------------------------------------------------------------------

function LibWeaponEnchant.RegisterCallback(token,func)
	if (not next(callbacks)) then
		f:RegisterEvent("PLAYER_ENTERING_WORLD");
		f:RegisterEvent("UNIT_INVENTORY_CHANGED");
		nextUpdate = 0.4;
		f:SetScript("OnUpdate",OnUpdate);
	end
	callbacks[token] = func;
end

function LibWeaponEnchant.UnregisterCallback(token)
	callbacks[token] = nil;
	if (not next(callbacks)) then
		f:UnregisterAllEvents();
	end
end

function LibWeaponEnchant.GetEnchantData()
	return mhSlot, mhEnchant, mhDuration, mhEndTime, ohSlot, ohEnchant, ohDuration, ohEndTime;
end